/*
 * (c) Stephen Chenney, University of Wisconsin at Madison, 2001
 *
 */

#ifndef _SPHEREWINDOW_H_
#define _SPHEREWINDOW_H_

#include <Fl/Fl.h>
#include <Fl/Fl_Gl_Window.h>
#include "Sphere.h"


// Subclass the Fl_Gl_Window because we want to draw OpenGL in here.
class SphereWindow : public Fl_Gl_Window {
    public:
	// Constructor takes window position and dimensions, the title.
	SphereWindow(int x, int y, int w, int h, char *label);

	// draw() method invoked whenever the view changes or the window
	// otherwise needs to be redrawn.
	void	draw(void);

	// Event handling method. Uses only mouse events.
	int	handle(int);

    private:
	Sphere	    sphere;
	bool	    smooth;

	static const double FOV_X; // The horizontal field of view.

	float	phi;	// Viewer's inclination angle.
	float	theta;	// Viewer's azimuthal angle.
	float	dist;	// Viewer's distance from the look-at point.
	int	x_last;
	int	y_last;

	int     button;	// The mouse button that is down, -1 if none.
	int	x_down; // The location of the mouse when the button was pushed
	int	y_down;
	float   phi_down;   // The view inclination angle when the mouse
			    // button was pushed
	float   theta_down; // The view azimuthal angle when the mouse
			    // button was pushed
	float	dist_down;  // The distance when the mouse button is pushed.

	void	Drag(void);
};


#endif

